from vcCommand import *
from math import *
import vcMatrix 
import os.path
app = getApplication()
cmd = getCommand()

SCRIPT_NAME = 'CABLE UPDATE SCRIPT'

def is_number(s):
  try:
    float(s)
    return True
  except ValueError:
    return False

def getAllNodes(parent, nodes = []):
  if not parent in nodes:
    nodes.append(parent)
  for node in parent.Children:
    getAllNodes(node, nodes)
  return nodes

def controllerChanged(arg):
  global jointName, controller, comp, activeController
  activeController = comp.findBehaviour(controller.Value)
  jointName.StepValues = []
  if (not activeController.JointCount > 0 or 
      not [x for x in activeController.Joints if x.Type == VC_JOINT_TRANSLATIONAL]):
    print 'No translational joints in the controller!'
    return
  l = []
  for joint in [x for x in activeController.Joints if x.Type == VC_JOINT_TRANSLATIONAL]:
    l.append(joint.Name)
  jointName.StepValues = l
  jointName.Value = jointName.StepValues[0]

def detailChanged(arg):
  if cableDetail.Value == 'Detailed':
    linksInHalfCircle.IsVisible = True
  else:
    linksInHalfCircle.IsVisible = False

def createCable(button):
  i = 1 +len([x for x in comp.Properties if 'Cable_' in x.Name and '::' in x.Name and 'CableWidth' in x.Name])
  cableIndex.Value = i
  
  if i == 0:
    cableName = name.Value
  else:
    cableName = '%s_%i'%(name.Value, cableIndex.Value)

  print 'creating cable with the name %s'%cableName

  ret = createCableProperties(cableName, test = True)
  if not ret:
    return None #something went wrong, check output for error decr.

  if cableDetail.Value == 'Simple':
    createSimpleCable(cableName)
  else:
    createDetailedCable(cableName)
  script = comp.findBehaviour(SCRIPT_NAME)
  if script:
    script.delete()
  script = comp.createBehaviour(VC_SCRIPT, SCRIPT_NAME)
  
  uri = getCommandPath()[8:]
  head, tail = os.path.split(uri)
  uri = os.path.join(head, 'EnergyCableWizard_temp_script.py')
  with open(uri,'r') as temp_script:
    content = temp_script.read()

  script.Script = content

  app.render()
  app.SelectionManager.setSelection(comp)

  i = 1 +len([x for x in comp.Properties if 'Cable_' in x.Name and '::' in x.Name and 'CableWidth' in x.Name])
  cableIndex.Value = i

  print 'created cable with the name %s'%cableName
  print 'To edit existing cables please refer to the properties (tabs) created for each cable'
  print 'i.e. DetailedCable_1::CableLength etc'

def createCableProperties(cableName, test = False, cableDetail = False):
  if cableLength.Value == '1500.0 || property => i.e. CableLength':
    print '''Please assign either a Value like 1500.0 or a name of an existing property \r
or an expression'''
    return None
  cableLength_prop = comp.createProperty(VC_EXPRESSION, '%s::%s'%(cableName, 'CableLength'))
  try:
    cableLength_prop.Value = cableLength.Value
  except:
    comp.deleteProperty(cableLength_prop.Name)
    print 'Please make sure that the variable used in the Cable Length (= %s) expression exists, and try again'%cableLength.Value
    return None
  if type(cableLength_prop.CalculatedValue) == str:
    comp.deleteProperty(cableLength_prop.Name)
    print 'Please make sure that the variable used in the Cable Length (= %s) expression exists, and try again'%cableLength.Value
    return None

  if test:
    comp.deleteProperty(cableLength_prop.Name)
    return True

  cableScale_prop = comp.createProperty(VC_REAL, '%s::%s'%(cableName, 'CableLengthScale'), VC_PROPERTY_LIMIT)
  cableScale_prop.MinValue = 0
  cableScale_prop.MaxValue = 1000
  cableScale_prop.Value = 1.0
  cableScale_prop.Quantity = 'Percentage'

  if not expressionProps.Value:
    radius_prop = comp.createProperty(VC_REAL, '%s::%s'%(cableName, 'CableCurveRadius'), VC_PROPERTY_LIMIT )
    radius_prop.MinValue = 10
    radius_prop.MaxValue = 1000
    radius_prop.Value = radius.Value
    radius_prop.Quantity = 'Distance'

    width_prop = comp.createProperty(VC_REAL, '%s::%s'%(cableName, 'CableWidth'), VC_PROPERTY_LIMIT )
    width_prop.MinValue = 1
    width_prop.MaxValue = 1000
    width_prop.Value = width.Value
    width_prop.Quantity = 'Distance'

    thickness_prop = comp.createProperty(VC_REAL, '%s::%s'%(cableName, 'CableThickness'), VC_PROPERTY_LIMIT )
    thickness_prop.MinValue = 1
    thickness_prop.MaxValue = 1000
    thickness_prop.Value = thickness.Value
    thickness_prop.Quantity = 'Distance'

  else:
    radius_prop = comp.createProperty(VC_EXPRESSION, '%s::%s'%(cableName, 'CableCurveRadius'))
    width_prop = comp.createProperty(VC_EXPRESSION, '%s::%s'%(cableName, 'CableWidth'))
    thickness_prop = comp.createProperty(VC_EXPRESSION, '%s::%s'%(cableName, 'CableThickness'))

  if cableDetail:
    linksInHalfCircle_prop = comp.createProperty(VC_INTEGER,  '%s::%s'%(cableName, 'LinksInHalfCircle'), VC_PROPERTY_LIMIT )
    linksInHalfCircle_prop.MinValue = 6
    linksInHalfCircle_prop.MaxValue = 30
    linksInHalfCircle_prop.Value = linksInHalfCircle.Value

    linkLength_prop = comp.createProperty(VC_REAL, '%s::%s'%(cableName, 'LinkLength'))
    linkLength_prop.WritableWhenConnected = False
    linkLength_prop.WritableWhenDisconnected = False
    linkLength_prop.WritableWhenSimulating = False
    linkLength_prop.Quantity = 'Distance'

    linkCount_prop = comp.createProperty(VC_INTEGER, '%s::%s'%(cableName, 'LinkCount'))
    linkCount_prop.WritableWhenConnected = False
    linkCount_prop.WritableWhenDisconnected = False
    linkCount_prop.WritableWhenSimulating = False

  jointName_prop = comp.createProperty(VC_STRING,'%s::%s'%(cableName, 'JointToFollow'), VC_PROPERTY_STEP )
  jointName_prop.StepValues = jointName.StepValues
  jointName_prop.Value = jointName.Value
  jointName_prop.WritableWhenSimulating = False

  cableColor_prop = comp.createProperty('Ref<Material>', '%s::%s'%(cableName, 'CableColor'))
  cableColor_prop.Value = cableColor.Value
  cableColor_prop.IsVisible = True

  update_prop = comp.createProperty(VC_BUTTON, '%s::%s'%(cableName, 'UpdateCable'))
  update_prop.ChangeOnRebuild = True

  if not expressionProps.Value:
    radius_prop.Value = radius.Value
    width_prop.Value = width.Value
    thickness_prop.Value = thickness.Value
  else:
    radius_prop.Value = '{}'.format(radius.Value)
    width_prop.Value = '{}'.format(width.Value)
    thickness_prop.Value = '{}'.format(thickness.Value)
  cableLength_prop.Value = cableLength.Value
  cableScale_prop.Value = cableScale.Value

  if cableDetail:
    return [cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, jointName_prop, cableColor_prop, linksInHalfCircle_prop, linkLength_prop, linkCount_prop]
  else:
    return [cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, jointName_prop, cableColor_prop]

def createSimpleCable(cableName):
  cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, jointName_prop, cableColor_prop = createCableProperties(cableName)
  nodeNames = ['%s'%cableName, '%s_Top'%cableName, '%s_Bot'%cableName, '%s_Curve'%cableName]
  nodes = []
  for i, nodeName in enumerate(nodeNames):
    node = comp.createNode(VC_NODE, nodeName)
    nodes.append(node)
    if i == 0:
      pass
    elif i == 1:
      nodes[0].attach(node)
      node.OffsetExpression = '%s(-%f)'%(cableDOF.Value, 10.0)
      transfrom = node.RootFeature.createFeature(VC_TRANSFORM, 'Box_transform')
      transfrom.Expression = 'Tz(%s)'%radius_prop.Name
      box = transfrom.createFeature(VC_BLOCK, 'Cable')
      box.Length = '10.0'
      box.Width = width_prop.Name
      box.Height = thickness_prop.Name
      box.Material = cableColor_prop.Value

      node.JointType = VC_JOINTTYPE_CUSTOM
      if len(cableDOF.Value) == 2:
        node.JointExpression = '%s(-%s).S%s(1.0+%s*.5/%f)'%(cableDOF.Value, jointName.Value, 
                                                 cableDOF.Value[1], jointName.Value, 10.0)
    elif i == 2:
      nodes[0].attach(node)
      node.OffsetExpression = '%s(-%s*%s).Tz(-%s-%s)'%(cableDOF.Value, cableLength_prop.Name, cableScale_prop.Name, radius_prop.Name, thickness_prop.Name)
      box = node.RootFeature.createFeature(VC_BLOCK, 'Cable')
      box.Length = '%s*%s'%(cableLength_prop.Name, cableScale_prop.Name)
      box.Width = width_prop.Name
      box.Height = thickness_prop.Name
      box.Material = cableColor_prop.Value

      node.JointType = VC_JOINTTYPE_CUSTOM
      if len(cableDOF.Value) == 2:
        node.JointExpression = 'S%s((%s*%s-%s*.5)/(%s*%s))'%(cableDOF.Value[1], cableLength_prop.Name, cableScale_prop.Name, jointName.Value, 
                                           cableLength_prop.Name, cableScale_prop.Name)
    elif i == 3:
      nodes[0].attach(node)
      revolve = node.RootFeature.createFeature(VC_REVOLVE, 'CableCurve')
      revolve.Sections = '36.0'
      revolve.RevolveAngle = '180.0'
      revolve.RevolveAxis = 'Vector(0, 1, 0)'
      revolve.Caps = True
      revolve.Material = cableColor_prop.Value

      transfrom = revolve.createFeature(VC_TRANSFORM, 'Box_transform_2')
      transfrom.Expression = 'Tz(%s)'%radius_prop.Name

      box = transfrom.createFeature(VC_BLOCK, 'Cable')
      box.Length = '10'
      box.Width = width_prop.Name
      box.Height = thickness_prop.Name
      box.Material = cableColor_prop.Value

      node.JointType = VC_JOINTTYPE_CUSTOM
      if len(cableDOF.Value) == 2:
        node.JointExpression = '%s(-%s*.5)'%(cableDOF.Value, jointName.Value)
  for node in nodes:
    node.rebuild()

def createDetailedCable(cableName):
  cableName = 'Detailed%s'%cableName
  
  props = createCableProperties(cableName, cableDetail = True)
  cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, \
  jointName_prop, cableColor_prop, linksInHalfCircle_prop, linkLength_prop, linkCount_prop = props

  node = comp.createNode(VC_NODE, cableName)
  node.JointType = VC_JOINTTYPE_CUSTOM
  node.JointExpression =  node.JointExpression = 'Tx(-%s)'%jointName.Value


  # = comp.createNode(VC_NODE, 'ChainFlange')
  driver = None
  linkLength_prop.Value = pi*radius_prop.Value / linksInHalfCircle_prop.Value
  linkCount_prop.Value = int(0.5+ (cableLength_prop.CalculatedValue*cableScale_prop.Value) / linkLength_prop.Value) + linksInHalfCircle_prop.Value
  for x in getAllNodes(comp):
    try:
      if x.Dof and x.Dof.Name == jointName_prop.Value:
        driver = x.Dof
        break
    except:
      pass
  n = node
  for i in range(linkCount_prop.Value):
    midpart = ((i%2)==0)
    name = "Link%d" % (i+1)
    n = buildNode(n, name, midpart, i, driver, props)
  node.rebuild()

def buildGeometry(node, props, midpart=False):
  cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, \
  jointName_prop, cableColor_prop, linksInHalfCircle_prop, linkLength_prop, linkCount_prop = props
  r = node.RootFeature

  for f in r.Children:
    f.delete()
  r.rebuild()

  t = r.createFeature(VC_TRANSFORM,'tb')
  if midpart:
    t.Expression = 'Tz(-0.5*%s).Ty(-0.4*%s)'%(thickness_prop.Name, width_prop.Name)
  else:
    t.Expression = 'Tz(-0.5*%s).Ty(-0.5*%s)'%(thickness_prop.Name, width_prop.Name)

  b = t.createFeature(VC_BLOCK,'b')
  b.Material = cableColor_prop.Value
  b.Length = '%s'%linkLength_prop.Name
  if midpart:
    b.Width = '0.8*%s'%width_prop.Name
  else:
    b.Width = '%s'%width_prop.Name
  b.Height = '%s'%thickness_prop.Name

  t = r.createFeature(VC_TRANSFORM,'tc1')
  if midpart:
    t.Expression = 'Rz(90).Ry(90).Tz(-0.4*%s)'%width_prop.Name
  else:
    t.Expression = 'Rz(90).Ry(90).Tz(-0.5*%s)'%width_prop.Name

  c = t.createFeature(VC_CYLINDER,'c1')
  c.Material = cableColor_prop.Value
  c.Radius = '0.5*%s'%thickness_prop.Name
  if midpart:
    c.Height = '0.8*%s'%width_prop.Name
  else:
    c.Height = '%s'%width_prop.Name
  c.EndSweep = '180'
  
  t = r.createFeature(VC_TRANSFORM,'tc2')
  if midpart:
    t.Expression = 'Rz(-90).Ry(-90).Tz(-0.4*%s).Ty(%s)'%(width_prop.Name, linkLength_prop.Name)
  else:
    t.Expression = 'Rz(-90).Ry(-90).Tz(-0.5*%s).Ty(%s)'%(width_prop.Name, linkLength_prop.Name)

  c = t.createFeature(VC_CYLINDER,'c1')
  c.Material = cableColor_prop.Value
  c.Radius = '0.5*%s'%thickness_prop.Name
  if midpart:
    c.Height = '0.8*%s'%width_prop.Name
  else:
    c.Height = '%s'%width_prop.Name
  c.EndSweep = '180'

  r.rebuild()
  
def buildNode(parent, name, midpart, index, driver, props):
  global ma
  cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, \
  jointName_prop, cableColor_prop, linksInHalfCircle_prop, linkLength_prop, linkCount_prop = props
  ll = linkLength_prop.Value
  ll2 = 2.0*ll
  rr = radius_prop.Value
  rr2 = 2.0*rr
  nd = index*ll # 'nominal' distance of node origin along chain
  rd = pi*rr # length of half circle perimeter
  ud = nd - rd # this would be the location on top half circle
  j1 = ud*2 # this is the joint value, where node origin is on top half circle
  j2 = nd*2 # this is the joint value, where node origin is on bottom half circle
  el = int(0.5+rd/ll) # number of links in half circle
  el = rd/ll # number of links in half circle
  ma = 180.0 / el # maximum angle between links in degrees
  
  # max stroke is 2 * the cable length - the length of the half circle
  sd = .5*cableLength_prop.CalculatedValue # max dist of the circle midpoint
  td = sd + rd # length of the cable (from previous)

  n = parent.createNode(VC_NODE,name)
  n.Offset = 'Tx(%s)'%linkLength_prop.Name
  n.JointType = VC_JOINTTYPE_ROTATIONAL_FOLLOWER
  n.Dof.AxisType = VC_JOINTAXIS_MINUS_Y
  n.Dof.Driver = driver

  # split expression and syntax to make easier to illustrate the idea

  # OUTSIDE AREA
  e1i = 'VALUE<%.2f||VALUE>%.2f' % (j1-ll,j2+ll) # outside the active area value is zero
  e1t = '0'
  
  # UPPER VARIABLE AREA
  e2i = 'VALUE<%.2f' % (j1+ll) # upper variable area in joint value
  e2t = 'asind(mod(0.5*(VALUE+%.2f),%.2f)/%.2f)' % (ll,ll,rr)
  
  # LOWER VARIABLE AREA
  e3i = 'VALUE>%.2f' % (j2-ll) # lower variable area
  e3t = '%.2f-asind(mod(0.5*(VALUE+%.2f),%.2f)/%.2f)' % (ma,ll,ll,rr)
  
  # INSIDE AREA
  e3e = '%.2f' % (ma) # inside the circle value is max angle
  
  e2e = '(%s?%s:%s)' % (e3i,e3t,e3e)
  e1e = '(%s?%s:%s)' % (e2i,e2t,e2e)

  expression = '%s?%s:%s' % (e1i,e1t,e1e)
  
  n.Dof.ValueExpression = expression
  buildGeometry(n, props, midpart)
  return n

def first_state():
  global name, jointName, radius, controller, comp, createCableProp, activeController, expressionProps
  global width, thickness, cableIndex, cableLength, cableDOF, cableScale, cableDetail, linksInHalfCircle, cableColor
  activeController = None
  comps = app.SelectionManager.getSelection(VC_SELECTION_COMPONENT)
  if not comps: 
    print 'No component selected, exiting!'
    return
  elif len(comps) > 1:
    print 'Select only one component, exiting!'
    return
  comp = comps[0]

  name = cmd.createProperty(VC_STRING, 'Cable Name' )
  name.Value = 'Cable'

  controller = cmd.createProperty(VC_STRING, 'Controller', VC_PROPERTY_STEP )
  controller.StepValues = []
  controller.OnChanged = controllerChanged

  jointName = cmd.createProperty(VC_STRING, 'Joint to follow', VC_PROPERTY_STEP )
  jointName.StepValues = []

  cableDetail = cmd.createProperty(VC_STRING, 'Visualization', VC_PROPERTY_STEP )
  cableDetail.StepValues = ['Simple', 'Detailed']
  cableDetail.Value = cableDetail.StepValues[0]
  cableDetail.OnChanged = detailChanged

  linksInHalfCircle = cmd.createProperty(VC_INTEGER, 'Links in half circle', VC_PROPERTY_LIMIT )
  linksInHalfCircle.MinValue = 6
  linksInHalfCircle.MaxValue = 30
  linksInHalfCircle.Value = 6
  linksInHalfCircle.IsVisible = False

  radius = cmd.createProperty(VC_REAL, 'Cable curve radius', VC_PROPERTY_LIMIT )
  radius.MinValue = 10
  radius.MaxValue = 1000
  radius.Value = 50
  radius.Quantity = 'Distance'

  width = cmd.createProperty(VC_REAL, 'Cable width', VC_PROPERTY_LIMIT )
  width.MinValue = 1
  width.MaxValue = 1000
  width.Value = 100
  width.Quantity = 'Distance'

  thickness = cmd.createProperty(VC_REAL, 'Cable thickness', VC_PROPERTY_LIMIT )
  thickness.MinValue = 1
  thickness.MaxValue = 1000
  thickness.Value = 50
  thickness.Quantity = 'Distance'

  cableIndex = cmd.createProperty(VC_INTEGER, 'Cable index', VC_PROPERTY_LIMIT )
  cableIndex.MinValue = 1
  cableIndex.MaxValue = 100
  cableIndex.Value = 1 +len([x for x in comp.Properties if 'Cable_' in x.Name and '::' in x.Name and 'CableWidth' in x.Name])
  cableIndex.WritableWhenConnected = False
  cableIndex.WritableWhenDisconnected = False
  cableIndex.WritableWhenSimulating = False

  expressionProps = cmd.createProperty(VC_BOOLEAN, 'Create Properties as Expressions')


  cableLength = cmd.createProperty(VC_STRING, 'Cable length')
  #cableLength.MinValue = 100
  #cableLength.MaxValue = 10000
  cableLength.Value = '1500.0 || property => i.e. CableLength'
  #cableLength.Quantity = 'Distance'

  cableScale = cmd.createProperty(VC_REAL, 'Cable length scale', VC_PROPERTY_LIMIT )
  cableScale.MinValue = 0
  cableScale.MaxValue = 1000
  cableScale.Value = 1.0
  cableScale.Quantity = 'Percentage'

  cableDOF = cmd.createProperty(VC_STRING, 'cableDOF', VC_PROPERTY_STEP )
  #cableDOF.StepValues = ['Tx', '-Tx', 'Ty', '-Ty', 'Tz', '-Tz']
  cableDOF.StepValues = ['Tx']
  cableDOF.Value = cableDOF.StepValues[0]
  cableDOF.IsVisible = False

  cableColor = cmd.createProperty('Ref<Material>', 'Cable Color')
  cableColor.IsVisible = True

  controllers = comp.findBehavioursByType(VC_SERVOCONTROLLER)
  controllers.extend(comp.findBehavioursByType(VC_ROBOTCONTROLLER))
  for _controller in controllers:
    l = controller.StepValues
    l.append(_controller.Name)
    controller.StepValues = l
  if not controllers:
    print 'No controllers detected, Servo- nor Robot Controllers, exiting!'
    return
  controller.Value = controller.StepValues[0]
  activeController = comp.findBehaviour(controller.Value)

  if (not activeController.JointCount > 0 or 
      not [x for x in activeController.Joints if x.Type == VC_JOINT_TRANSLATIONAL]):
    print 'No translational joints in the controller!'
  else:
    l = []
    for joint in [x for x in activeController.Joints if x.Type == VC_JOINT_TRANSLATIONAL]:
      l.append(joint.Name)
    jointName.StepValues = l
    jointName.Value = jointName.StepValues[0]

  createCableProp = cmd.createProperty(VC_BUTTON, 'Create cable' )
  createCableProp.OnChanged = createCable

  #print VC_JOINTTYPE_CUSTOM
  #print VC_JOINTTYPE_FIXED
  #print VC_JOINTTYPE_ROTATIONAL
  #print VC_JOINTTYPE_ROTATIONAL_FOLLOWER
  #print VC_JOINTTYPE_TRANSLATIONAL
  #print VC_JOINTTYPE_TRANSLATIONAL_FOLLOWER 
  #print VC_JOINT_ROTATIONAL 
  #print VC_JOINT_TRANSLATIONAL 
  executeInActionPanel()

addState( first_state )