from vcScript import *
from math import * 
from vcHelpers.Selection import getGivenNodesFeatures as getGivenNodesFeatures
import vcMatrix 
app = getApplication()
comp = getComponent()

def getAllNodes(parent, nodes = []):
  if not parent in nodes:
    nodes.append(parent)
  for node in parent.Children:
    getAllNodes(node, nodes)
  return nodes

def updateCable(arg):
  cableName = arg.Name.split('::')[0]
  if 'Detailed' in cableName:
    createDetailedCable(cableName)
  else:
    parent = comp.findNode(cableName)
    mat = comp.getProperty('%s::%s'%(cableName, 'CableColor')).Value
    for node in getAllNodes(parent):
      for fea in getGivenNodesFeatures(node):
        if fea.getProperty('Material'):
          fea.Material = mat
    parent.rebuild()
    app.render()

def getCableProperties(cableName):
  propNames = ['%s::%s'%(cableName, 'CableLength'), '%s::%s'%(cableName, 'CableCurveRadius'), \
               '%s::%s'%(cableName, 'CableWidth'), '%s::%s'%(cableName, 'CableThickness'), \
               '%s::%s'%(cableName, 'CableLengthScale'), '%s::%s'%(cableName, 'JointToFollow'), \
                '%s::%s'%(cableName, 'CableColor'), '%s::%s'%(cableName, 'LinksInHalfCircle'), \
               '%s::%s'%(cableName, 'LinkLength'), '%s::%s'%(cableName, 'LinkCount')]

  return map(comp.getProperty, propNames)

def createDetailedCable(cableName):
  node = comp.findNode(cableName)
  for c in node.Children:
      c.delete()

  props = getCableProperties(cableName)
  cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, \
  jointName_prop, cableColor_prop, linksInHalfCircle_prop, linkLength_prop, linkCount_prop = props

  # = comp.createNode(VC_NODE, 'ChainFlange')
  driver = None
  linkLength_prop.Value = pi*radius_prop.Value / linksInHalfCircle_prop.Value
  linkCount_prop.Value = int(0.5+ (cableLength_prop.CalculatedValue*cableScale_prop.Value) / linkLength_prop.Value) + linksInHalfCircle_prop.Value
  for x in getAllNodes(comp):
    try:
      if x.Dof and x.Dof.Name == jointName_prop.Value:
        driver = x.Dof
        break
    except:
      pass
  n = node
  for i in range(linkCount_prop.Value):
    midpart = ((i%2)==0)
    name = "Link%d" % (i+1)
    n = buildNode(n, name, midpart, i, driver, props)
  node.rebuild()

def buildGeometry(node, props, midpart=False):
  cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, \
  jointName_prop, cableColor_prop, linksInHalfCircle_prop, linkLength_prop, linkCount_prop = props
  r = node.RootFeature

  for f in r.Children:
    f.delete()
  r.rebuild()

  t = r.createFeature(VC_TRANSFORM,'tb')
  if midpart:
    t.Expression = 'Tz(-0.5*%s).Ty(-0.4*%s)'%(thickness_prop.Name, width_prop.Name)
  else:
    t.Expression = 'Tz(-0.5*%s).Ty(-0.5*%s)'%(thickness_prop.Name, width_prop.Name)

  b = t.createFeature(VC_BLOCK,'b')
  b.Material = cableColor_prop.Value
  b.Length = '%s'%linkLength_prop.Name
  if midpart:
    b.Width = '0.8*%s'%width_prop.Name
  else:
    b.Width = '%s'%width_prop.Name
  b.Height = '%s'%thickness_prop.Name

  t = r.createFeature(VC_TRANSFORM,'tc1')
  if midpart:
    t.Expression = 'Rz(90).Ry(90).Tz(-0.4*%s)'%width_prop.Name
  else:
    t.Expression = 'Rz(90).Ry(90).Tz(-0.5*%s)'%width_prop.Name

  c = t.createFeature(VC_CYLINDER,'c1')
  c.Material = cableColor_prop.Value
  c.Radius = '0.5*%s'%thickness_prop.Name
  if midpart:
    c.Height = '0.8*%s'%width_prop.Name
  else:
    c.Height = '%s'%width_prop.Name
  c.EndSweep = '180'
  
  t = r.createFeature(VC_TRANSFORM,'tc2')
  if midpart:
    t.Expression = 'Rz(-90).Ry(-90).Tz(-0.4*%s).Ty(%s)'%(width_prop.Name, linkLength_prop.Name)
  else:
    t.Expression = 'Rz(-90).Ry(-90).Tz(-0.5*%s).Ty(%s)'%(width_prop.Name, linkLength_prop.Name)

  c = t.createFeature(VC_CYLINDER,'c1')
  c.Material = cableColor_prop.Value
  c.Radius = '0.5*%s'%thickness_prop.Name
  if midpart:
    c.Height = '0.8*%s'%width_prop.Name
  else:
    c.Height = '%s'%width_prop.Name
  c.EndSweep = '180'

  r.rebuild()
  
def buildNode(parent, name, midpart, index, driver, props):
  global ma
  cableLength_prop, radius_prop, width_prop, thickness_prop, cableScale_prop, \
  jointName_prop, cableColor_prop, linksInHalfCircle_prop, linkLength_prop, linkCount_prop = props
  ll = linkLength_prop.Value
  ll2 = 2.0*ll
  rr = radius_prop.Value
  rr2 = 2.0*rr
  nd = index*ll # 'nominal' distance of node origin along chain
  rd = pi*rr # length of half circle perimeter
  ud = nd - rd # this would be the location on top half circle
  j1 = ud*2 # this is the joint value, where node origin is on top half circle
  j2 = nd*2 # this is the joint value, where node origin is on bottom half circle
  el = int(0.5+rd/ll) # number of links in half circle
  el = rd/ll # number of links in half circle
  ma = 180.0 / el # maximum angle between links in degrees
  
  # max stroke is 2 * the cable length - the length of the half circle
  sd = .5*cableLength_prop.CalculatedValue # max dist of the circle midpoint
  td = sd + rd # length of the cable (from previous)

  n = parent.createNode(VC_NODE,name)
  n.Offset = 'Tx(%s)'%linkLength_prop.Name
  n.JointType = VC_JOINTTYPE_ROTATIONAL_FOLLOWER
  n.Dof.AxisType = VC_JOINTAXIS_MINUS_Y
  n.Dof.Driver = driver

  # split expression and syntax to make easier to illustrate the idea

  # OUTSIDE AREA
  e1i = 'VALUE<%.2f||VALUE>%.2f' % (j1-ll,j2+ll) # outside the active area value is zero
  e1t = '0'
  
  # UPPER VARIABLE AREA
  e2i = 'VALUE<%.2f' % (j1+ll) # upper variable area in joint value
  e2t = 'asind(mod(0.5*(VALUE+%.2f),%.2f)/%.2f)' % (ll,ll,rr)
  
  # LOWER VARIABLE AREA
  e3i = 'VALUE>%.2f' % (j2-ll) # lower variable area
  e3t = '%.2f-asind(mod(0.5*(VALUE+%.2f),%.2f)/%.2f)' % (ma,ll,ll,rr)
  
  # INSIDE AREA
  e3e = '%.2f' % (ma) # inside the circle value is max angle
  
  e2e = '(%s?%s:%s)' % (e3i,e3t,e3e)
  e1e = '(%s?%s:%s)' % (e2i,e2t,e2e)

  expression = '%s?%s:%s' % (e1i,e1t,e1e)
  
  n.Dof.ValueExpression = expression
  buildGeometry(n, props, midpart)
  return n 

props = []
for p in [x for x in comp.Properties if 'UpdateCable' in x.Name]:
  props.append(p)
  p.OnChanged = updateCable
